#include "shared.h"
#include "dinput.h"
#include "vfw.h"
#include "stdio.h"
#include "resource.h"
#include "file.h"
#include "audio.h"
#include "input.h"
#include "main.h"
#include "registry.h"

static BMP			Bmp;
static HWND			hwnd;
static HINSTANCE	hInstance;
static HMENU		hMenu;
static BOOL			RunEmulation = FALSE;
static BOOL			Paused = FALSE;
static HANDLE		Mutex = NULL;
static unsigned int	USER_MESSAGE = 0;
unsigned char		BitmapData[256 * 192];

static char			SRAMname[MAX_PATH] = "";
static unsigned int	Frame		  = 0;
static unsigned int	LastTickCount = 0;

static unsigned int	FrameTime60hz[60] = {17, 17, 16, 17, 17, 16, 17, 17, 16, 17, 17, 16, 17, 17, 16,
										 17, 17, 16, 17, 17, 16, 17, 17, 16, 17, 17, 16, 17, 17, 16,
										 17, 17, 16, 17, 17, 16, 17, 17, 16, 17, 17, 16, 17, 17, 16,
										 17, 17, 16, 17, 17, 16, 17, 17, 16, 17, 17, 16, 17, 17, 16};

static unsigned int	FrameTime50hz[50] = {20, 20, 20, 20, 20, 20, 20, 20, 20, 20,
										 20, 20, 20, 20, 20, 20, 20, 20, 20, 20,
										 20, 20, 20, 20, 20, 20, 20, 20, 20, 20,
										 20, 20, 20, 20, 20, 20, 20, 20, 20, 20,
										 20, 20, 20, 20, 20, 20, 20, 20, 20, 20};

int WINAPI WinMain(HINSTANCE hinstance, HINSTANCE hPrevInstance, PSTR szCmdLine, int iCmdShow)
{
	WNDCLASS    WindowClass;
	MSG			msg;
	int			QuitFlag = FALSE;

	if (CreateSingleInstance(szCmdLine))
	{
		// Clear the window class structure
		memset(&WindowClass, 0, sizeof(WindowClass));

		// Set up window class attributes
		WindowClass.style                   = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;	// Window styles
		WindowClass.lpfnWndProc				= MainWindowProc;						// Windows messaging function
		WindowClass.hInstance				= hInstance;							// Application instance
		WindowClass.hIcon                   = LoadIcon(NULL, IDI_APPLICATION);		// Icon to display on title bar
		WindowClass.hCursor                 = LoadCursor(NULL, IDC_ARROW);			// Cursor to use over window
		WindowClass.hbrBackground			= (HBRUSH)(COLOR_APPWORKSPACE + 1);		// Colour to use for window background
		WindowClass.lpszMenuName			= NULL;									// Menu to use
		WindowClass.lpszClassName			= SMS_CLASS_NAME;						// Class name

		// Register window class
		if (!RegisterClass(&WindowClass))
		{
			// Return unsuccessful
			return FALSE;
		}

		InitRegistryEntries();
		LoadRegistryEntries();

		// Create the main window for our application
		hInstance = hinstance;
		hwnd = CreateWindowEx(0,									// Extended window style
							  SMS_CLASS_NAME,			            // Pointer to registered class name
							  "SMS Plus",							// Pointer to window name 
							  WS_POPUP | WS_SYSMENU | WS_CAPTION | WS_MINIMIZEBOX,	// Window style
							  0,                                    // X Pos
							  0,                                    // Y Pos
							  0,									// Width
							  0,							        // Height
							  NULL,                                 // Handle to parent or owner window
							  NULL,                                 // Handle to menu, or child window
							  hInstance,					        // Application instance
							  NULL);                                // Pointer to window creation data

		// Load the menu to use for the application and set it
		hMenu = LoadMenu(hInstance, MAKEINTRESOURCE(IDR_MENU));
		SetMenu(hwnd, hMenu);

		SetScreenScale(RegistryInfo.ScreenScale);
		SetUpdateFrequency(RegistryInfo.UpdateFrequency);
		SetFrameSkip(RegistryInfo.FrameSkip);
		SetSoundFrequency(RegistryInfo.SoundFrequency);
		ShowWindow(hwnd, SW_SHOW);

		InitVRAM();
		InitAudio(hwnd, RegistryInfo.SoundFrequency, FREQ_60HZ);
		InitInput(hwnd, hInstance);

		if (strlen(szCmdLine))
		{
			if (!stricmp(&szCmdLine[strlen(szCmdLine) - 2], "gg") ||
				!stricmp(&szCmdLine[strlen(szCmdLine) - 3], "sms"))
			{
				OpenROM(szCmdLine);
			}
		}
	}
	else
	{
		QuitFlag = TRUE;
	}

	// Main program loop
    while (QuitFlag == FALSE)
    {
        if (PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE))
        {
			if (GetMessage(&msg, NULL, 0, 0))
            {
                TranslateMessage(&msg);
                DispatchMessage(&msg);
		    }
            else
            {
                // GetMessage returning FALSE implies exit message
                QuitFlag = TRUE;
            }
        }
        else
        {
			unsigned int		TickCount;
			unsigned int		*pFrameTime;
			unsigned int		Skip = (Frame % (RegistryInfo.FrameSkip + 1))? 1 : 0;

			if (RegistryInfo.UpdateFrequency == FREQ_50HZ)
			{
				pFrameTime = FrameTime50hz;
			}
			else
			{
				pFrameTime = FrameTime60hz;
			}

			TickCount = GetTickCount();

			if (LastTickCount == 0) LastTickCount = TickCount;

			if (TickCount - LastTickCount >= pFrameTime[Frame % RegistryInfo.UpdateFrequency])
			{
				if (RunEmulation && !Paused)
				{
			        sms_frame(Skip);

					if (RegistryInfo.EnableSound) UpdateAudio(snd.buffer, snd.bufsize, Frame++);

					if (!Skip)
					{
						UpdatePalette();
						UpdateKeyboard();
						UpdateSpecialKeys();
						UpdateJoysticks();
						UpdateRapidFire(Frame);
						PostMessage(hwnd, WM_PAINT, 0, 0);
					}
				}

				LastTickCount = TickCount - ((TickCount - LastTickCount) - pFrameTime[Frame % RegistryInfo.UpdateFrequency]);

				if ((TickCount - LastTickCount) > (pFrameTime[0] * NO_OF_AUDIO_BUFFERS))
				{
					Frame		  = 0;
					LastTickCount = 0;
				}
		    }
		}
    }

	SaveRegistryEntries();
	TidyInput();
	TidyAudio();

	if (hMenu) DestroyMenu(hMenu);
	DestroySingleInstance();

	return msg.wParam;
}

LRESULT CALLBACK MainWindowProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	static HBITMAP		hBitmap = NULL;
	
	switch (msg)
	{
	case WM_CREATE:
		{
			HDC			hDC;

			memset(&Bmp, 0, sizeof(BMP));
			Bmp.BitmapInfoHeader.biSize			= sizeof(BITMAPINFOHEADER);
			Bmp.BitmapInfoHeader.biWidth		= 256;
			Bmp.BitmapInfoHeader.biHeight		= 192;
			Bmp.BitmapInfoHeader.biPlanes		= 1;
			Bmp.BitmapInfoHeader.biBitCount		= 8;
			Bmp.BitmapInfoHeader.biCompression	= BI_RGB;
			Bmp.BitmapInfoHeader.biClrUsed		= 256;

			hDC = GetDC(hwnd);
			hBitmap = CreateDIBitmap(hDC, &Bmp.BitmapInfoHeader, 0, &BitmapData[0], (BITMAPINFO *)&Bmp, DIB_PAL_COLORS);
			ReleaseDC(hwnd, hDC);

			SendMessage(hwnd, WM_COMMAND, ID_MENU_CONFIG_SOUND_ENABLE, 0);
		}
		break;

	case WM_ACTIVATEAPP:
		if (wParam)
		{
			Paused = FALSE;

			// Restore Sync (+ Restart Audio)
			Frame		  = 0;
			LastTickCount = 0;

			// Restore Input
			InitInput(hwnd, hInstance);
		}
		else
		{
			Paused = TRUE;
			StopAudio();
			TidyInput();
		}
		break;

	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case ID_MENU_FILE_OPEN:
		    SaveSRAM(SRAMname);
			OpenROM(NULL);
			break;

		case ID_MENU_FILE_EXIT:
			SendMessage(hwnd, WM_CLOSE, 0, 0);
			break;

		case ID_MENU_CPU_RESET:
			input_mask |= KEY_PAD_RESET;
			break;

		case ID_MENU_CONFIG_SCREEN_X1:
			SetScreenScale(RegistryInfo.ScreenScale = 1);
			break;

		case ID_MENU_CONFIG_SCREEN_X2:
			SetScreenScale(RegistryInfo.ScreenScale = 2);
			break;

		case ID_MENU_CONFIG_SCREEN_X3:
			SetScreenScale(RegistryInfo.ScreenScale = 3);
			break;

		case ID_MENU_CONFIG_UPDATE_50HZ:
			SetUpdateFrequency(RegistryInfo.UpdateFrequency = FREQ_50HZ);
			break;

		case ID_MENU_CONFIG_UPDATE_60HZ:
			SetUpdateFrequency(RegistryInfo.UpdateFrequency = FREQ_60HZ);
			break;

		case ID_MENU_CONFIG_SKIP_NO:
			SetFrameSkip(RegistryInfo.FrameSkip = 0);
			break;

		case ID_MENU_CONFIG_SKIP_1:
			SetFrameSkip(RegistryInfo.FrameSkip = 1);
			break;

		case ID_MENU_CONFIG_SKIP_2:
			SetFrameSkip(RegistryInfo.FrameSkip = 2);
			break;

		case ID_MENU_CONFIG_SKIP_3:
			SetFrameSkip(RegistryInfo.FrameSkip = 3);
			break;

		case ID_MENU_CONFIG_SKIP_4:
			SetFrameSkip(RegistryInfo.FrameSkip = 4);
			break;
		
		case ID_MENU_CONFIG_SOUND_ENABLE:
			if (GetMenuState(hMenu, ID_MENU_CONFIG_SOUND_ENABLE, MF_BYCOMMAND) == MF_CHECKED)
			{
				RegistryInfo.EnableSound = FALSE;
				CheckMenuItem(hMenu, ID_MENU_CONFIG_SOUND_ENABLE, MF_UNCHECKED);
			}
			else
			{
				RegistryInfo.EnableSound = TRUE;
				CheckMenuItem(hMenu, ID_MENU_CONFIG_SOUND_ENABLE, MF_CHECKED);
			}
			break;

		case ID_MENU_CONFIG_SOUND_44100:
			SetSoundFrequency(RegistryInfo.SoundFrequency = 44100);
			break;

		case ID_MENU_CONFIG_SOUND_22050:
			// Pick nearest divisible by 50 & 60
			SetSoundFrequency(RegistryInfo.SoundFrequency = 22200);
			break;

		case ID_MENU_CONFIG_SOUND_11025:
			// Pick nearest divisible by 50 & 60
			SetSoundFrequency(RegistryInfo.SoundFrequency = 11100);
			break;

		case ID_MENU_CONFIG_CONTROLS:
			DialogBox(hInstance, MAKEINTRESOURCE(IDD_CONTROLS), hwnd, ConfigureControlsProc);
			TidyInput();
			InitInput(hwnd, hInstance);
			break;

		case ID_MENU_CONFIG_FOLDERS:
			DialogBox(hInstance, MAKEINTRESOURCE(IDD_FOLDERS), hwnd, ConfigureFoldersProc);
			break;

		case ID_MENU_CONFIG_ASSOCIATIONS:
			if (MessageBox(hwnd, "Associate the file types '.gg' and '.sms' with SMS Plus?", "", MB_YESNO) == IDYES)
			{
				SetFileAssociations();
				MessageBox(hwnd, "File associations set.", "", MB_OK);
			}
			break;

		case ID_MENU_HELP_ABOUT:
			DialogBox(hInstance, MAKEINTRESOURCE(IDD_ABOUT), hwnd, AboutProc);
			break;

		default:
			return TRUE;
		}
		break;

	case WM_ENTERMENULOOP:
		StopAudio();
		break;

	case WM_EXITMENULOOP:
		Frame		  = 0;
		LastTickCount = 0;
		break;

	case WM_PAINT:
		{
			HDC					hDC;
			HDRAWDIB			hDrawDib;
			PAINTSTRUCT			PaintStruct;

			BeginPaint(hwnd, &PaintStruct);
			EndPaint(hwnd, &PaintStruct);

			hDC = GetDC(hwnd);
			hDrawDib = DrawDibOpen();
			if (cart.type == TYPE_GG)
			{
				DrawDibBegin(hDrawDib, hDC, Bmp.BitmapInfoHeader.biWidth - 96, Bmp.BitmapInfoHeader.biHeight - 48, &Bmp.BitmapInfoHeader, 
							 Bmp.BitmapInfoHeader.biWidth, Bmp.BitmapInfoHeader.biHeight, DDF_BACKGROUNDPAL);
				DrawDibDraw(hDrawDib, hDC, 0, 0, (Bmp.BitmapInfoHeader.biWidth - 96) * RegistryInfo.ScreenScale, (Bmp.BitmapInfoHeader.biHeight - 48) * RegistryInfo.ScreenScale,
							&Bmp.BitmapInfoHeader, &BitmapData[0], 48, 24, Bmp.BitmapInfoHeader.biWidth - 96, Bmp.BitmapInfoHeader.biHeight - 48, DDF_BACKGROUNDPAL);
			}
			else
			{
				DrawDibBegin(hDrawDib, hDC, Bmp.BitmapInfoHeader.biWidth, Bmp.BitmapInfoHeader.biHeight, &Bmp.BitmapInfoHeader, 
							 Bmp.BitmapInfoHeader.biWidth, Bmp.BitmapInfoHeader.biHeight, DDF_BACKGROUNDPAL);
				DrawDibDraw(hDrawDib, hDC, 0, 0, Bmp.BitmapInfoHeader.biWidth * RegistryInfo.ScreenScale, Bmp.BitmapInfoHeader.biHeight * RegistryInfo.ScreenScale,
							&Bmp.BitmapInfoHeader, &BitmapData[0], 0, 0, Bmp.BitmapInfoHeader.biWidth, Bmp.BitmapInfoHeader.biHeight, DDF_BACKGROUNDPAL);
			}
			DrawDibEnd(hDrawDib);
			DrawDibClose(hDrawDib);
			ReleaseDC(hwnd, hDC);
		}
		break;

	case WM_CLOSE:
		DestroyWindow(hwnd);
		break;

	case WM_DESTROY:
	    SaveSRAM(SRAMname);
        PostQuitMessage(0);
		break;

	default:
		if (msg == USER_MESSAGE)
		{
		    SaveSRAM(SRAMname);
			OpenROM((char *)lParam);
			break;
		}
		else
		{
			return DefWindowProc(hwnd, msg, wParam, lParam);
		}
	}

	return FALSE;
}

void AdjustWindowSize(void)
{
	if (cart.type == TYPE_GG)
	{
		RECT	Rect  = {0, 0, 160 * RegistryInfo.ScreenScale, 144 * RegistryInfo.ScreenScale};

		AdjustWindowRect(&Rect, GetWindowLong(hwnd, GWL_STYLE), GetMenu(hwnd)? TRUE : FALSE);
		SetWindowPos(hwnd, NULL, 0, 0, Rect.right - Rect.left, Rect.bottom - Rect.top, 
					 SWP_NOMOVE | SWP_NOZORDER | SWP_NOACTIVATE | SWP_FRAMECHANGED);
	}
	else
	{
		RECT	Rect  = {0, 0, 256 * RegistryInfo.ScreenScale, 192 * RegistryInfo.ScreenScale};

		AdjustWindowRect(&Rect, GetWindowLong(hwnd, GWL_STYLE), GetMenu(hwnd)? TRUE : FALSE);
		SetWindowPos(hwnd, NULL, 0, 0, Rect.right - Rect.left, Rect.bottom - Rect.top, 
					 SWP_NOMOVE | SWP_NOZORDER | SWP_NOACTIVATE | SWP_FRAMECHANGED);
	}
}

void InitVRAM(void)
{
     /* Initialize display bitmap */
    memset(&bitmap, 0, sizeof(t_bitmap));
    bitmap.width  = 256;
    bitmap.height = 192;
    bitmap.depth  = 8;
    bitmap.pitch  = (bitmap.width * (bitmap.depth >> 3));
    bitmap.data   = (unsigned char *)&BitmapData[0];
}

void UpdatePalette(void)
{
    unsigned char	r, g, b;
    int				index;

    if (bitmap.pal.update == 0) return;

    for (index = 0; index < 32; index++)
    {
        if(bitmap.pal.color[index])
        {
            r = bitmap.pal.color[index][0];
            g = bitmap.pal.color[index][1];
            b = bitmap.pal.color[index][2];

			Bmp.Palette[0x00 | index].rgbRed   = r;
			Bmp.Palette[0x00 | index].rgbGreen = g;
			Bmp.Palette[0x00 | index].rgbBlue  = b;
			Bmp.Palette[0x20 | index].rgbRed   = r;
			Bmp.Palette[0x20 | index].rgbGreen = g;
			Bmp.Palette[0x20 | index].rgbBlue  = b;
			Bmp.Palette[0x40 | index].rgbRed   = r;
			Bmp.Palette[0x40 | index].rgbGreen = g;
			Bmp.Palette[0x40 | index].rgbBlue  = b;
        }
    }

    bitmap.pal.update = 0;
    memset(bitmap.pal.dirty, 0, 32);
}

void OpenROM(char *pROM)
{
	HANDLE			FileHandle;
	char			Filename[MAX_PATH] = "";
	int				FileSize;
	int				BytesRead;
	int				i;

	if (pROM)
	{
		strcpy(Filename, pROM);
	}
	else
	{
		// Initialise the file dialog
		InitFileDialog(hwnd, RegistryInfo.ROMPath, "Roms (*.gg, *.sms)\0*.gg;*.sms\0All Files (*.*)\0*.*\0", 0); 

		// Show the dialog, and return if it is cancelled
		if (!GetLoadFileFromDialog(hwnd, "Open", Filename)) return;
	}

	FileHandle = CreateFile(Filename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);

	if (FileHandle == INVALID_HANDLE_VALUE) return;
	
    FileSize = GetFileSize(FileHandle, NULL);

	if (FileSize < 0x8000) 
	{
		CloseHandle(FileHandle);
		return;
	}

	if ((FileSize / 512) & 1)
	{
		SetFilePointer(FileHandle, 512, 0, FILE_CURRENT);
		FileSize -= 512;
	}

	if (!stricmp(&Filename[strlen(Filename) - 2], "gg"))
	{
	    cart.type  = TYPE_GG;
	}
	else
	{
	    cart.type  = TYPE_SMS;
	}

	cart.pages = (FileSize / 0x4000);
    
	if (cart.rom) free(cart.rom);
	cart.rom = malloc(FileSize);
    
	if (!cart.rom)
	{
		CloseHandle(FileHandle);
		return;
	}

    ReadFile(FileHandle, cart.rom, FileSize, &BytesRead, NULL); 

    CloseHandle(FileHandle);

	input_mask |= KEY_PAD_RESET;
	system_reset();
	system_init(RegistryInfo.SoundFrequency, RegistryInfo.UpdateFrequency);

	for (i = strlen(Filename) - 1 ; i >= 0 ; i--)
	{
		if (Filename[i] == '\\') 
		{
			break;
		}
	}

    strcpy(SRAMname, &Filename[i + 1]);
    strcpy(strrchr(SRAMname, '.'), ".sav\0");
    LoadSRAM(SRAMname);

	AdjustWindowSize();
	RunEmulation = TRUE;
}

char *GetSRAMName(void)
{
	return SRAMname;
}

char *GetBitmapInfoHeader(void)
{
	return (char *)&Bmp;
}

char *GetBitmapData(void)
{
	return BitmapData;
}

void LoadSRAM(char *pSRAMname)
{
	HANDLE		FileHandle;
	int			BytesRead;
	char		Filename[MAX_PATH];

	wsprintf(Filename, "%s%s", RegistryInfo.SRAMPath, pSRAMname);

    FileHandle = CreateFile(Filename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);

	if (FileHandle == INVALID_HANDLE_VALUE) return;

    ReadFile(FileHandle, sms.sram, 0x8000, &BytesRead, NULL); 

	CloseHandle(FileHandle);

	sms.save = 1;
}

void SaveSRAM(char *pSRAMname)
{
    HANDLE		FileHandle;
    int			BytesWritten;
	char		Filename[MAX_PATH];

	wsprintf(Filename, "%s%s", RegistryInfo.SRAMPath, pSRAMname);

	if (sms.save && strlen(pSRAMname))
	{
	    FileHandle = CreateFile(Filename, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_FLAG_SEQUENTIAL_SCAN, NULL);

	    if (FileHandle == INVALID_HANDLE_VALUE) return;

	    WriteFile(FileHandle, sms.sram, 0x8000, &BytesWritten, NULL); 

	    CloseHandle (FileHandle);
	}
}

LRESULT CALLBACK ConfigureControlsProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg)
	{
	case WM_INITDIALOG:
		{
			RECT		ParentRect;
			RECT		OurRect;
			POINT		Pos;

			// Center the window
			GetClientRect(GetParent(hwnd), &ParentRect);
			GetClientRect(hwnd, &OurRect);
			Pos.x = (ParentRect.right - OurRect.right) / 2;
			Pos.y = (ParentRect.bottom - OurRect.bottom) / 2;
			ClientToScreen(GetParent(hwnd), &Pos);
			SetWindowPos(hwnd, HWND_TOP, Pos.x, Pos.y, 0, 0, SWP_NOSIZE);

			PopulateJoystickCombo(hwnd, PLAYER_1);
			PopulateJoystickCombo(hwnd, PLAYER_2);
			PopulateConfigureControls(hwnd);

			CheckDlgButton(hwnd, IDC_P1_BUTTON1_RAPIDFIRE_CHECK, RegistryInfo.RapidFire[PLAYER_1][PLAYER1_BUTTON1 - PLAYER1_BUTTON1]? BST_CHECKED : BST_UNCHECKED);
			CheckDlgButton(hwnd, IDC_P1_BUTTON2_RAPIDFIRE_CHECK, RegistryInfo.RapidFire[PLAYER_1][PLAYER1_BUTTON2 - PLAYER1_BUTTON1]? BST_CHECKED : BST_UNCHECKED);
			CheckDlgButton(hwnd, IDC_P2_BUTTON1_RAPIDFIRE_CHECK, RegistryInfo.RapidFire[PLAYER_2][PLAYER2_BUTTON1 - PLAYER2_BUTTON1]? BST_CHECKED : BST_UNCHECKED);
			CheckDlgButton(hwnd, IDC_P2_BUTTON2_RAPIDFIRE_CHECK, RegistryInfo.RapidFire[PLAYER_2][PLAYER2_BUTTON2 - PLAYER2_BUTTON1]? BST_CHECKED : BST_UNCHECKED);

			ShowWindow(hwnd, SW_SHOW);
		}
		break;

	case WM_COMMAND:
		{
			switch (HIWORD(wParam))
			{
			case BN_CLICKED:
				switch (LOWORD (wParam))
				{
				case IDC_P1_LEFT_BUTTON:
					ConfigureButton(hwnd, PLAYER1_LEFT, IDC_P1_LEFT, "'Player 1 Left'");
					break;

				case IDC_P1_RIGHT_BUTTON:
					ConfigureButton(hwnd, PLAYER1_RIGHT, IDC_P1_RIGHT, "'Player 1 Right'");
					break;

				case IDC_P1_UP_BUTTON:
					ConfigureButton(hwnd, PLAYER1_UP, IDC_P1_UP, "'Player 1 Up'");
					break;

				case IDC_P1_DOWN_BUTTON:
					ConfigureButton(hwnd, PLAYER1_DOWN, IDC_P1_DOWN, "'Player 1 Down'");
					break;

				case IDC_P1_BUTTON1_BUTTON:
					ConfigureButton(hwnd, PLAYER1_BUTTON1, IDC_P1_BUTTON1, "'Player 1 Button 1'");
					break;

				case IDC_P1_BUTTON2_BUTTON:
					ConfigureButton(hwnd, PLAYER1_BUTTON2, IDC_P1_BUTTON2, "'Player 1 Button 2'");
					break;

				case IDC_P1_START_BUTTON:
					ConfigureButton(hwnd, PLAYER1_START, IDC_P1_START, "'Player 1 Start / Pause'");
					break;

				case IDC_P2_LEFT_BUTTON:
					ConfigureButton(hwnd, PLAYER2_LEFT, IDC_P2_LEFT, "'Player 2 Left'");
					break;

				case IDC_P2_RIGHT_BUTTON:
					ConfigureButton(hwnd, PLAYER2_RIGHT, IDC_P2_RIGHT, "'Player 2 Right'");
					break;

				case IDC_P2_UP_BUTTON:
					ConfigureButton(hwnd, PLAYER2_UP, IDC_P2_UP_BUTTON, "'Player 2 Up'");
					break;

				case IDC_P2_DOWN_BUTTON:
					ConfigureButton(hwnd, PLAYER2_DOWN, IDC_P2_DOWN, "'Player 2 Down'");
					break;

				case IDC_P2_BUTTON1_BUTTON:
					ConfigureButton(hwnd, PLAYER2_BUTTON1, IDC_P2_BUTTON1, "'Player 2 Button 1'");
					break;

				case IDC_P2_BUTTON2_BUTTON:
					ConfigureButton(hwnd, PLAYER2_BUTTON2, IDC_P2_BUTTON2, "'Player 2 Button 2'");
					break;

				case IDC_P2_START_BUTTON:
					ConfigureButton(hwnd, PLAYER2_START, IDC_P2_START, "'Player 2 Start / Pause'");
					break;

				case IDC_P1_BUTTON1_RAPIDFIRE_CHECK:
					RegistryInfo.RapidFire[PLAYER_1][PLAYER1_BUTTON1 - PLAYER1_BUTTON1] ^= 1;
					break;

				case IDC_P1_BUTTON2_RAPIDFIRE_CHECK:
					RegistryInfo.RapidFire[PLAYER_1][PLAYER1_BUTTON2 - PLAYER1_BUTTON1] ^= 1;
					break;
				
				case IDC_P2_BUTTON1_RAPIDFIRE_CHECK:
					RegistryInfo.RapidFire[PLAYER_2][PLAYER2_BUTTON1 - PLAYER2_BUTTON1] ^= 1;
					break;
				
				case IDC_P2_BUTTON2_RAPIDFIRE_CHECK:
					RegistryInfo.RapidFire[PLAYER_2][PLAYER2_BUTTON2 - PLAYER2_BUTTON1] ^= 1;
					break;
				}
				break;

			case CBN_SELCHANGE:
				{
					int		Index = SendMessage(GetDlgItem(hwnd, LOWORD(wParam)), CB_GETCURSEL, 0, 0);
	
					switch (LOWORD(wParam))
					{
					case IDC_P1_JOYSTICK_COMBO:
						SetSelectedJoystick(PLAYER_1, Index);
						PopulateConfigureControls(hwnd);
						break;

					case IDC_P2_JOYSTICK_COMBO:
						SetSelectedJoystick(PLAYER_2, Index);
						PopulateConfigureControls(hwnd);
						break;
					}
				}
				break;
			}
		}
		break;

	case WM_CLOSE:
		EndDialog(hwnd, 0);
		break;

    default:
		return FALSE;
	}

	return TRUE;
}

void PopulateConfigureControls(HWND hwnd)
{
	int		Joystick1 = GetSelectedJoystick(PLAYER_1);
	int		Joystick2 = GetSelectedJoystick(PLAYER_2);
	char	Text[32];

	if (Joystick1)
	{
		EnableWindow(GetDlgItem(hwnd, IDC_P1_LEFT_BUTTON), FALSE);
		EnableWindow(GetDlgItem(hwnd, IDC_P1_RIGHT_BUTTON), FALSE);
		EnableWindow(GetDlgItem(hwnd, IDC_P1_UP_BUTTON), FALSE);
		EnableWindow(GetDlgItem(hwnd, IDC_P1_DOWN_BUTTON), FALSE);

		SetDlgItemText(hwnd, IDC_P1_LEFT, 	 "Left");
		SetDlgItemText(hwnd, IDC_P1_RIGHT,	 "Right");
		SetDlgItemText(hwnd, IDC_P1_UP,		 "Up");
		SetDlgItemText(hwnd, IDC_P1_DOWN,	 "Down");

		wsprintf(Text, "Button %d", RegistryInfo.JoystickMap[0][PLAYER1_BUTTON1 - PLAYER1_BUTTON1]);
		SetDlgItemText(hwnd, IDC_P1_BUTTON1, Text);
		wsprintf(Text, "Button %d", RegistryInfo.JoystickMap[0][PLAYER1_BUTTON2 - PLAYER1_BUTTON1]);
		SetDlgItemText(hwnd, IDC_P1_BUTTON2, Text);
		wsprintf(Text, "Button %d", RegistryInfo.JoystickMap[0][PLAYER1_START - PLAYER1_BUTTON1]);
		SetDlgItemText(hwnd, IDC_P1_START, Text);
	}
	else
	{
		EnableWindow(GetDlgItem(hwnd, IDC_P1_LEFT_BUTTON), TRUE);
		EnableWindow(GetDlgItem(hwnd, IDC_P1_RIGHT_BUTTON), TRUE);
		EnableWindow(GetDlgItem(hwnd, IDC_P1_UP_BUTTON), TRUE);
		EnableWindow(GetDlgItem(hwnd, IDC_P1_DOWN_BUTTON), TRUE);

		SetDlgItemText(hwnd, IDC_P1_LEFT,	 ButtonToString(PLAYER1_LEFT));
		SetDlgItemText(hwnd, IDC_P1_RIGHT,	 ButtonToString(PLAYER1_RIGHT));
		SetDlgItemText(hwnd, IDC_P1_UP,		 ButtonToString(PLAYER1_UP));
		SetDlgItemText(hwnd, IDC_P1_DOWN,	 ButtonToString(PLAYER1_DOWN));
		SetDlgItemText(hwnd, IDC_P1_BUTTON1, ButtonToString(PLAYER1_BUTTON1));
		SetDlgItemText(hwnd, IDC_P1_BUTTON2, ButtonToString(PLAYER1_BUTTON2));
		SetDlgItemText(hwnd, IDC_P1_START,	 ButtonToString(PLAYER1_START));
	}
	
	if (Joystick2)
	{
		EnableWindow(GetDlgItem(hwnd, IDC_P2_LEFT_BUTTON), FALSE);
		EnableWindow(GetDlgItem(hwnd, IDC_P2_RIGHT_BUTTON), FALSE);
		EnableWindow(GetDlgItem(hwnd, IDC_P2_UP_BUTTON), FALSE);
		EnableWindow(GetDlgItem(hwnd, IDC_P2_DOWN_BUTTON), FALSE);

		SetDlgItemText(hwnd, IDC_P2_LEFT, 	 "Left");
		SetDlgItemText(hwnd, IDC_P2_RIGHT,	 "Right");
		SetDlgItemText(hwnd, IDC_P2_UP,		 "Up");
		SetDlgItemText(hwnd, IDC_P2_DOWN,	 "Down");

		wsprintf(Text, "Button %d", RegistryInfo.JoystickMap[PLAYER_2][PLAYER2_BUTTON1 - PLAYER2_BUTTON1]);
		SetDlgItemText(hwnd, IDC_P2_BUTTON1, Text);
		wsprintf(Text, "Button %d", RegistryInfo.JoystickMap[PLAYER_2][PLAYER2_BUTTON2 - PLAYER2_BUTTON1]);
		SetDlgItemText(hwnd, IDC_P2_BUTTON2, Text);
		wsprintf(Text, "Button %d", RegistryInfo.JoystickMap[PLAYER_2][PLAYER2_START - PLAYER2_BUTTON1]);
		SetDlgItemText(hwnd, IDC_P2_START, Text);
	}
	else
	{
		EnableWindow(GetDlgItem(hwnd, IDC_P2_LEFT_BUTTON), TRUE);
		EnableWindow(GetDlgItem(hwnd, IDC_P2_RIGHT_BUTTON), TRUE);
		EnableWindow(GetDlgItem(hwnd, IDC_P2_UP_BUTTON), TRUE);
		EnableWindow(GetDlgItem(hwnd, IDC_P2_DOWN_BUTTON), TRUE);

		SetDlgItemText(hwnd, IDC_P2_LEFT,	 ButtonToString(PLAYER2_LEFT));
		SetDlgItemText(hwnd, IDC_P2_RIGHT,	 ButtonToString(PLAYER2_RIGHT));
		SetDlgItemText(hwnd, IDC_P2_UP,		 ButtonToString(PLAYER2_UP));
		SetDlgItemText(hwnd, IDC_P2_DOWN,	 ButtonToString(PLAYER2_DOWN));
		SetDlgItemText(hwnd, IDC_P2_BUTTON1, ButtonToString(PLAYER2_BUTTON1));
		SetDlgItemText(hwnd, IDC_P2_BUTTON2, ButtonToString(PLAYER2_BUTTON2));
		SetDlgItemText(hwnd, IDC_P2_START,	 ButtonToString(PLAYER2_START));
	}
}

void PopulateJoystickCombo(HWND hwnd, int Player)
{
	int		i;

	if (Player == PLAYER_1)
	{
		hwnd = GetDlgItem(hwnd, IDC_P1_JOYSTICK_COMBO);
	}
	else
	{
		hwnd = GetDlgItem(hwnd, IDC_P2_JOYSTICK_COMBO);
	}

	SendMessage(hwnd, CB_RESETCONTENT, 0, 0);
	SendMessage(hwnd, CB_ADDSTRING, 0, (long)"None");

	for (i = 0 ; i < MAX_JOYSTICKS ; i++)
	{
		char	Name[MAX_PATH];

		if (GetJoystickName(i, Name))
		{
			SendMessage(hwnd, CB_ADDSTRING, 0, (long)Name);
		}
	}

	SendMessage(hwnd, CB_SETCURSEL, GetSelectedJoystick(Player), 0);
}

int ConfigureButton(HWND hwnd, int Button, int DialogItem, char *pString)
{
	int			AssignedKey;

	AssignedKey = DialogBoxParam(hInstance, MAKEINTRESOURCE(IDD_CONTROLS_KEY), hwnd, ConfigureControlsKeyProc, (long)pString);
	SetControlKey(Button, AssignedKey);
	
	if (AssignedKey >= 256)
	{
		char		Text[32];

		wsprintf(Text, "Button %d", AssignedKey - 256);
		SetDlgItemText(hwnd, DialogItem, Text);
	}
	else
	{
		SetDlgItemText(hwnd, DialogItem, ButtonToString(Button));
	}

	return AssignedKey;
}

LRESULT CALLBACK ConfigureControlsKeyProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	static	UINT_PTR	Timer = 0;

	switch (msg)
	{
	case WM_INITDIALOG:
		{
			RECT		ParentRect;
			RECT		OurRect;
			POINT		Pos;
			char		String[MAX_PATH];

			// Center the window
			GetClientRect(GetParent(hwnd), &ParentRect);
			GetClientRect(hwnd, &OurRect);
			Pos.x = (ParentRect.right - OurRect.right) / 2;
			Pos.y = (ParentRect.bottom - OurRect.bottom) / 2;
			ClientToScreen(GetParent(hwnd), &Pos);
			SetWindowPos(hwnd, HWND_TOP, Pos.x, Pos.y, 0, 0, SWP_NOSIZE);
			ShowWindow(hwnd, SW_SHOW);

			if (JoystickPresent())
			{
				wsprintf(String, "Press the new key or button for %s.", (char *)lParam);
			}
			else
			{
				wsprintf(String, "Press the new key for %s.", (char *)lParam);
			}
			SetDlgItemText(hwnd, IDC_TEXT, String);

			TidyInput();
			InitInput(hwnd, hInstance);
			Timer = SetTimer(hwnd, 0, 20, NULL);
		}
		break;

	case WM_TIMER:
		{
			int		Key = GetKeyPressed(TRUE);

			if (Key == -1 && JoystickPresent())
			{
				UpdateJoysticks();
				Key = GetJoystickButtonPressed();
			}

			if (Key != -1)
			{
				KillTimer(hwnd, Timer);
				EndDialog(hwnd, Key);
			}
		}
		break;

    default:
		return FALSE;
	}

	return TRUE;
}

void SetScreenScale(unsigned int Scale)
{
	CheckMenuItem(hMenu, ID_MENU_CONFIG_SCREEN_X1, (Scale == 1)? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(hMenu, ID_MENU_CONFIG_SCREEN_X2, (Scale == 2)? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(hMenu, ID_MENU_CONFIG_SCREEN_X3, (Scale == 3)? MF_CHECKED : MF_UNCHECKED);
	AdjustWindowSize();
}

void SetUpdateFrequency(unsigned int Frequency)
{
	CheckMenuItem(hMenu, ID_MENU_CONFIG_UPDATE_50HZ, (Frequency == FREQ_50HZ)? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(hMenu, ID_MENU_CONFIG_UPDATE_60HZ, (Frequency == FREQ_60HZ)? MF_CHECKED : MF_UNCHECKED);
	TidyAudio();
	audio_init(RegistryInfo.SoundFrequency, Frequency);
	InitAudio(hwnd, RegistryInfo.SoundFrequency, Frequency);
}

void SetFrameSkip(unsigned int Skip)
{
	CheckMenuItem(hMenu, ID_MENU_CONFIG_SKIP_NO, (Skip == 0)? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(hMenu, ID_MENU_CONFIG_SKIP_1, (Skip == 1)? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(hMenu, ID_MENU_CONFIG_SKIP_2, (Skip == 2)? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(hMenu, ID_MENU_CONFIG_SKIP_3, (Skip == 3)? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(hMenu, ID_MENU_CONFIG_SKIP_4, (Skip == 4)? MF_CHECKED : MF_UNCHECKED);
}

void SetSoundFrequency(unsigned int Frequency)
{
	CheckMenuItem(hMenu, ID_MENU_CONFIG_SOUND_44100, (Frequency == 44100)? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(hMenu, ID_MENU_CONFIG_SOUND_22050, (Frequency == 22200)? MF_CHECKED : MF_UNCHECKED);
	CheckMenuItem(hMenu, ID_MENU_CONFIG_SOUND_11025, (Frequency == 11100)? MF_CHECKED : MF_UNCHECKED);
	TidyAudio();
	audio_init(Frequency, RegistryInfo.UpdateFrequency);
	InitAudio(hwnd, Frequency, RegistryInfo.UpdateFrequency);
}

LRESULT CALLBACK ConfigureFoldersProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg)
	{
	case WM_INITDIALOG:
		{
			RECT		ParentRect;
			RECT		OurRect;
			POINT		Pos;

			// Center the window
			GetClientRect(GetParent(hwnd), &ParentRect);
			GetClientRect(hwnd, &OurRect);
			Pos.x = (ParentRect.right - OurRect.right) / 2;
			Pos.y = (ParentRect.bottom - OurRect.bottom) / 2;
			ClientToScreen(GetParent(hwnd), &Pos);
			SetWindowPos(hwnd, HWND_TOP, Pos.x, Pos.y, 0, 0, SWP_NOSIZE);

			SetDlgItemText(hwnd, IDC_ROM_FOLDER, RegistryInfo.ROMPath);
			SetDlgItemText(hwnd, IDC_SRAM_FOLDER, RegistryInfo.SRAMPath);
			SetDlgItemText(hwnd, IDC_SCREENSHOT_FOLDER, RegistryInfo.ScreenshotPath);

			ShowWindow(hwnd, SW_SHOW);
		}
		break;

	case WM_COMMAND:
		{
			switch (HIWORD(wParam))
			{
			case BN_CLICKED:
				switch (LOWORD (wParam))
				{
				case IDC_ROM_BROWSE:
					if (BrowseForDirectory(hwnd, RegistryInfo.ROMPath, "Select ROM folder"))
					{
						SetDlgItemText(hwnd, IDC_ROM_FOLDER, RegistryInfo.ROMPath);
					}
					break;

				case IDC_SRAM_BROWSE:
					if (BrowseForDirectory(hwnd, RegistryInfo.SRAMPath, "Select SRAM folder"))
					{
						if (RegistryInfo.SRAMPath[strlen(RegistryInfo.SRAMPath) - 1] != '\\')
						{
							strcat(RegistryInfo.SRAMPath, "\\");
						}
						SetDlgItemText(hwnd, IDC_SRAM_FOLDER, RegistryInfo.SRAMPath);
					}
					break;

				case IDC_SCREENSHOT_BROWSE:
					if (BrowseForDirectory(hwnd, RegistryInfo.ScreenshotPath, "Select Screenshot folder"))
					{
						if (RegistryInfo.ScreenshotPath[strlen(RegistryInfo.ScreenshotPath) - 1] != '\\')
						{
							strcat(RegistryInfo.ScreenshotPath, "\\");
						}
						SetDlgItemText(hwnd, IDC_SCREENSHOT_FOLDER, RegistryInfo.ScreenshotPath);
					}
					break;

				case IDC_OK:
					SendMessage(hwnd, WM_CLOSE, 0, 0);
				}
				break;
			}
		}
		break;

	case WM_CLOSE:
		EndDialog(hwnd, 0);
		break;

    default:
		return FALSE;
	}

	return TRUE;
}

LRESULT CALLBACK AboutProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg)
	{
	case WM_INITDIALOG:
		{
			RECT		ParentRect;
			RECT		OurRect;
			POINT		Pos;

			// Center the window
			GetClientRect(GetParent(hwnd), &ParentRect);
			GetClientRect(hwnd, &OurRect);
			Pos.x = (ParentRect.right - OurRect.right) / 2;
			Pos.y = (ParentRect.bottom - OurRect.bottom) / 2;
			ClientToScreen(GetParent(hwnd), &Pos);
			SetWindowPos(hwnd, HWND_TOP, Pos.x, Pos.y, 0, 0, SWP_NOSIZE);

			ShowWindow(hwnd, SW_SHOW);
		}
		break;

	case WM_CLOSE:
		EndDialog(hwnd, 0);
		break;

    default:
		return FALSE;
	}

	return TRUE;
}

BOOL CreateSingleInstance(char *pCommandLine)
{
	SECURITY_ATTRIBUTES			Security;

	USER_MESSAGE = RegisterWindowMessage(SMS_INSTANCE_LOAD_NAME);

	Security.nLength			  = sizeof(SECURITY_ATTRIBUTES);
	Security.lpSecurityDescriptor = NULL;
	Security.bInheritHandle		  = FALSE;

	Mutex = CreateMutex(&Security, TRUE, SMS_MUTEX_NAME);

	if (Mutex)
	{
		int			State = WaitForSingleObject(Mutex, 0);

		if (State != WAIT_TIMEOUT) return TRUE;
	}

	SendMessage(FindWindow(SMS_CLASS_NAME, NULL), USER_MESSAGE, 0, (long)pCommandLine);

	DestroySingleInstance();

	return FALSE;
}

void DestroySingleInstance(void)
{
	if (Mutex)
	{
		CloseHandle(Mutex);
		Mutex = NULL;
	}
}

BOOL IsWindowsNT4(void)
{
	OSVERSIONINFO		Info;

	Info.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);

	GetVersionEx(&Info);

	if ((Info.dwMajorVersion == 4) && (Info.dwPlatformId == VER_PLATFORM_WIN32_NT))
	{
		return TRUE;
	}
	else
	{
		return FALSE;
	}
}